/*
 * cvsadmin.c - cvs users administration program
 *
 * Copyright (c) 2001, Vincent Labrecque <limitln@cooptel.qc.ca>
 * All rights reserved.
 * 
 *      Redistribution and use in source and binary forms,
 *      with or without modification, are permitted provided
 *      that the following conditions are met:
 * 
 *      Redistributions of source code must retain the
 *      above copyright notice, this list of conditions
 *      and the following disclaimer. 
 *      Redistributions in binary form must reproduce
 *      the above copyright notice, this list of
 *      conditions and the following disclaimer in the
 *      documentation and/or other materials provided
 *      with the distribution. 
 *      Neither the name of the <ORGANIZATION> nor the
 *      names of its contributors may be used to
 *      endorse or promote products derived from this
 *      software without specific prior written
 *      permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS
 * AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
 * SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include "backend.h"
#include "util.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>

#ifdef HAVE_CONFIG_H
# include "../config.h"
#endif

#ifdef HAVE_CRYPT_H
#include <crypt.h>
#endif
#ifdef HAVE_GETOPT_H
# include <getopt.h>
#endif


#define DEFAULT_CVSROOT ":local:/usr/local/cvsroot"

#define CHOMP(s)	do {			\
	if((s)[strlen((s))-1] == '\n')		\
		(s)[strlen((s))-1] = '\0';	\
	} while(0)

int id;				/* current database ID */
char *repos;			/* repository name */
bool usecvs;			/* do we use cvs(1) to get our files? */
bool precomputed_hash;		/* use precomputed hashes instead of passwords? */


int ask_yes_no(char *fmt, ...)
{
	va_list args;
	int c;
	char buf[256];

	/* 
	 * Print a buffer, so printf() doesn't have to do its stuff all the
	 * time.. (this limits the size of your questions, but well ;))
	 */
	va_start(args, fmt);
	vsnprintf(buf, sizeof(buf), fmt, args);

	for (;;) {
		printf("%s", buf);
		fflush(stdout);
		c = getchar();
		if (c == EOF) return 0;
		if (c == 'y' || c == 'y') return 1;
		if (c == 'n' || c == 'N') return 0;

		printf("Invalid character, choose [yYnN]\n");
	}
	va_end(args);

	return 0;
}

/*
 * Prompt the user for "something".
 *
 * if hidden is set we call getpass and return the crypt()ed input.
 */
char *do_prompt(bool hidden, char *fmt, ...)
{
	char buf[BUFSIZ], *s, *p, salt[3];
	va_list args;


	s = NULL;

	va_start(args, fmt);
	vfprintf(stdout, fmt, args); fflush(stdout);
	va_end(args);

	/*
	 * Getpass stores input in a static buffer, so we copy it, and then
	 * clear it. There's a window when it's wide open, and even after
	 * this it's still recoverable, but well, I think this is still
	 * better than nothing.
	 */
	if (hidden) {
		generate_salt(salt); 
		p = getpass("");
		/* s = strdup(precomputed_hash?p:crypt(p, salt)); */
		s = strdup(crypt(p, salt));
		while (*p != '\0')
			*p++ = '\0';
	}
	else {
		if (fgets(buf, sizeof(buf), stdin) == NULL) {
			perror("fgets");
			exit(1);
		}
		CHOMP(buf);
		if(!strlen(buf))
			s = NULL;
		else
			s = strdup(buf);
	}

	return s;
}

/*****************************     handlers ******/

int list_handler(int argc, char **argv)
{
	char *pass, *sys, *email;
	int i, x;

	if(argc > 1) {
		for(i = 1; i < argc; i++) {
			x = cvsadmin_get_user_info(id,
				argv[i], &pass, &sys, &email);
			if(x < 0)
				printf("%s: no such user\n", argv[1]);
			else
				printf("%s<%s>: %s on sys\n",
					argv[1], email?email:"no@email", sys?sys:argv[1]);
		}
	} else {
		char **list;

		list = cvsadmin_get_user_list(id);
		if(list==NULL)
			printf("Can't get users list\n");
		else{
			for(i = 0; list[i]!=NULL; i++) {
				x = cvsadmin_get_user_info(id,
					list[i], &pass, &sys, &email);
				/*
				 * This means that cvsadmin returned an
				 * invalid username in the first place, so
				 * just exit before we break everything...
				 */
				if(x < 0) {
					fprintf(stderr, "Internal fsck up in cvsadmin!\n");
					exit(1);
				}
				printf("%s<%s>: %s on sys\n",
					list[i], email?email:"no@email", sys);
			}
			cvsadmin_free_str_list(list);
		}
	}
	return 0;
}

int email_handler(int argc, char **argv)
{
	int i;
	char *email;
	
	for(i = 1; i < argc; i++) {
		email = do_prompt(0, "New email for %s? ", argv[i]);
		if(cvsadmin_set_user_info(id, argv[i], NULL, NULL, email) < 0)
			printf("Error setting email to %s\n", email);
		/* 
		 * It's copied in cvsadmin_set_user_info so we can safely free it
		 */
		free(email);
	}
	return 0;
}

int rename_handler(int argc, char **argv)
{
	int i;
	char *old, *new;

	for(i = 1; i < argc-1; i+=2) {
		old = argv[i];
		new = argv[i+1];
		if(cvsadmin_rename_user(id, old, new) < 0)
			printf("Error %s->%s\n", old, new);
	}
	return 0;
}

static void prompt_user_info(char *uname, char **pass, char **sysuser, char **email)
{
	printf("User %s\n", uname);
	if(precomputed_hash)
		*pass = do_prompt(0, "Password hash? ");
	else
		*pass = do_prompt(1, "Password? ");
	*sysuser = do_prompt(0, "System user? [%s] ", *sysuser?*sysuser:"");
	*email = do_prompt(0, "Email address? [%s] ", *email?*email:"");
}

int misc_handler(int argc, char **argv)
{
	int i;

	for(i = 1; i < argc; i++) {
		char *uname, *pass, *sysuser, *email;

		uname = argv[i];
		if(cvsadmin_get_user_info(id, uname, &pass,
						&sysuser, &email) < 0){
			printf("error getting %s's info\n",uname);
			continue;
		}
		prompt_user_info(uname, &pass, &sysuser, &email);
		if(cvsadmin_set_user_info(id, uname, pass, sysuser, email) < 0)
			printf("Error setting %s's new info\n",uname);
	}

	return 0;
}

int add_handler(int argc, char **argv)
{
	int i;

	for(i = 1; i < argc; i++) {
		char *uname, *pass, *sysuser, *email;

		uname = argv[i];
		pass = sysuser = email = NULL;

		prompt_user_info(uname, &pass, &sysuser, &email);

		if(cvsadmin_set_user_info(id, uname, pass, sysuser, email) < 0)
			printf("Error setting %s's new info\n",uname);
	}

	return 0;
}

int del_handler(int argc, char **argv)
{
	int i;

	for(i = 1; i < argc; i++){
		if(cvsadmin_del_user(id, argv[i]) < 0)
			printf("Error deleting %s\n", argv[i]);
	}

	return 0;
}

int sys_handler(int argc, char **argv)
{
	int i;
	char *sys;

	for(i = 1; i < argc; i++) {
		if(cvsadmin_get_user_info(id, argv[i], NULL, &sys, NULL) < 0){
			printf("Can't get %s's information\n", argv[i]);
			continue;
		}
		sys = do_prompt(0, "New system user for %s? [%s] ? ",
				argv[i], sys);
		if(sys != NULL) {
			if(cvsadmin_set_user_info(id, argv[i], NULL, sys, NULL) < 0){
				printf("Error changing %s's system user\n",
					argv[i]);
			}
		}
	}

	return 0;
}

int pass_handler(int argc, char **argv)
{
	int i;
	char *opass, *npass;

	for(i = 1; i < argc; i++) {
		if(cvsadmin_get_user_info(id, argv[i], &opass, NULL, NULL) < 0) {
			printf("Can't get %s's information\n", argv[i]);
			continue;
		}
		npass = do_prompt(1, "New password for %s? ", argv[i]);
		if(npass == NULL) {
			printf("No password is bad. Really. Not changing\n");
		} else {
			if(cvsadmin_set_user_info(id, argv[i], npass, NULL, NULL) < 0)
				printf("Error setting %s's new pass\n", argv[i]);
		}
	}

	return 0;
}

struct {
	char *name;
	int (*handler)(int argc, char **argv);
	char *parameters;	/* for usage()...*/
} actions[] = {
	{ "list", list_handler,  "list of usernames || nothing" },
	{ "add", add_handler, "list of usernames" },
	{ "del", del_handler, "list of usernames" },
	{ "email", email_handler,"user list" },
	{ "rename", rename_handler, "list of alternating olduser/newuser" },
	{ "misc", misc_handler, "list of usernames" },
	{ "sys", sys_handler, "list of usernames" },
	{ "pass", pass_handler, "list of usernames" },
	{ NULL, 0x0 },
};

void usage()
{
	int i;
	extern char *__progname;

	printf("Usage: \n%s [options] [action]\n\n", __progname);
	printf("options:\n");
	printf("\t-c - Use cvs(1) to checkout the password/users files\n");
	printf("\t-C - Ask for precomputed hashes instead of passwords\n");
	printf("\t-h - Get this help\n");
	printf("\t-d [repository] - Set the repository's location\n");
	printf("\t-v - Show current version of cvsadmin\n");
	printf("actions:\n");	
	printf("\tName\tArguments\n");
	printf("\t----\t---------\n");
	for (i = 0; actions[i].name != NULL; i++)
		printf("\t%s\t%s\n", actions[i].name, actions[i].parameters);

	exit(1);
}


int main(int argc, char *argv[])
{
	char *action;
	int i, ret, ch;

	while ((ch = getopt(argc,argv,"cCd:vh")) != -1)
		switch(ch) {
		case 'c':
			usecvs = 1;
			break;
		case 'C':
			precomputed_hash = 1;
			break;
		case 'd':
			repos = strdup(optarg);
			break;
		case 'v':
			printf("%s %s\n", PACKAGE, VERSION);
			exit(0);
		default:
			usage();
			/* NOTREACHED */
		}
	action = argv[optind];
	if (action == NULL)
		usage();

	for(i = 0; actions[i].name != NULL; i++) {
		if(!strcmp(action, actions[i].name))
			break;
	}
	if(actions[i].name == NULL) {
		printf("Invalid action %s\n", action);
		usage();
	}

	if(!repos) {
		repos = getenv("CVSROOT");
		if(!repos)
			repos = DEFAULT_CVSROOT;
	}
	id = cvsadmin_open(repos, usecvs);
	if(id < 0) {
		printf("Unable to open %s\n", repos);
		exit(1);
	}

	ret = actions[i].handler(argc - optind, &argv[optind]);

	if (cvsadmin_close(id) < 0)
		printf("Unable to close repository\n");

	return ret;
}
